import { system } from "@minecraft/server";
import { saveEntityLocation, removeEntityLocation, chestEntities } from "./index";


export class ChestUtils {
    constructor(chest_entity) {
        this.chest_entity = chest_entity;
        this.maxOpen = 5;
        this.maxClose = -1;
        this.openInterval = 0;
        this.closeInterval = 0;
        this.cooldown = 1;
    }
    open() {
        const { chest_entity, maxOpen, cooldown } = this;
        const { dimension, location } = chest_entity;
        if (chest_entity.hasTag('custom_open')) {
            const chest = this.getChest(this.chest_entity);
            if (chest.hasTag('custom_chest')) {
                const isDouble = chest.permutation.getState('aga_iron_chest:top') === 1 || chest.permutation.getState('aga_iron_chest:bottom') === 1;
                const chests = isDouble ? this.getDoubleChests(chest) : [chest];
    
                dimension.playSound('random.chestopen', location);
                chest_entity.setDynamicProperty('isOpen', true);
    
                this.openInterval = system.runInterval(() => {
                    chests.forEach(c => {
                        const chestState = c.permutation.getState('aga_iron_chest:open');
                        const nextChestState = c.permutation.withState('aga_iron_chest:open', chestState + 1);
                        c.setPermutation(nextChestState);
                    });
                    if ((chests[0].permutation.getState('aga_iron_chest:open') + 1) >= maxOpen) {
                        system.clearRun(this.openInterval);
                        this.startClosingInterval();
                    }
                }, cooldown);
            }
            else {
                // Barrel logic remains unchanged
                dimension.playSound('block.barrel.open', location);
                chest_entity.setDynamicProperty('isOpen', true);
                const openBarrel = chest.permutation.withState('aga_iron_chest:open', 1);
                chest.setPermutation(openBarrel);
                this.startClosingInterval();
            }
        }
    }

    startClosingInterval() {
        this.closeInterval = system.runInterval(() => {
            this.close();
        }, this.cooldown);
    }

    close() {
        const { chest_entity, maxClose } = this;
        const { dimension, location } = chest_entity;
        if (chest_entity.hasTag('custom_open')) return;

        const chest = this.getChest(chest_entity);
        if (chest?.hasTag('custom_chest')) {
            const isDouble = chest.permutation.getState('aga_iron_chest:top') === 1 || chest.permutation.getState('aga_iron_chest:bottom') === 1;
            const chests = isDouble ? this.getDoubleChests(chest) : [chest];
            const chestState = chests[0].permutation.getState('aga_iron_chest:open');
            
            if ((chestState - 1) <= maxClose) {
                if (chest_entity.getDynamicProperty('isOpen')) {
                    chest_entity.setDynamicProperty('isOpen', false);
                    dimension.playSound('random.chestclosed', location);
                }
                system.clearRun(this.closeInterval);
                return;
            }

            chests.forEach(c => {
                const nextChestState = c.permutation.withState('aga_iron_chest:open', chestState - 1);
                c.setPermutation(nextChestState);
            });
        }
        else {
            // Barrel logic
            if (chest_entity.getDynamicProperty('isOpen')) {
                chest_entity.setDynamicProperty('isOpen', false);
                const closedBarrel = chest.permutation.withState('aga_iron_chest:open', 0);
                chest.setPermutation(closedBarrel);
                dimension.playSound('block.barrel.close', location);
                system.clearRun(this.closeInterval);
            }
        }
    }

    
    // Helper method to get both chests of a vertical double chest
    getDoubleChests(chest) {
        const isTop = chest.permutation.getState('aga_iron_chest:top') === 1;
        const isBottom = chest.permutation.getState('aga_iron_chest:bottom') === 1;
        if (!isTop && !isBottom) return [chest];

        const { dimension } = this.chest_entity;
        const chestDirection = chest.permutation.getState('minecraft:cardinal_direction');

        // Check vertically instead of horizontally
        const checkDirections = [
            { dy: 1, direction: 'up' },
            { dy: -1, direction: 'down' }
        ];

        for (const dir of checkDirections) {
            const adjacentLocation = { 
                x: chest.location.x, 
                y: chest.location.y + dir.dy, 
                z: chest.location.z 
            };
            const adjacentBlock = dimension.getBlock(adjacentLocation);

            if (adjacentBlock && adjacentBlock.typeId === chest.typeId) {
                const adjacentDirection = adjacentBlock.permutation.getState('minecraft:cardinal_direction');
                if (adjacentDirection !== chestDirection) continue; // Skip if orientations don't match

                const adjacentIsTop = adjacentBlock.permutation.getState('aga_iron_chest:top') === 1;
                const adjacentIsBottom = adjacentBlock.permutation.getState('aga_iron_chest:bottom') === 1;
                if ((isTop && adjacentIsBottom) || (isBottom && adjacentIsTop)) {
                    return [chest, adjacentBlock];
                }
            }
        }

        return [chest]; // Fallback to single chest if no valid partner is found
    }
    drop() {
        const { chest_entity } = this;
        const { dimension, location } = chest_entity;
        const isDouble = chest_entity.typeId.includes('_double');
    
        if (isDouble) {
            const inv = chest_entity.getComponent('inventory').container;
            const halfSize = Math.floor(inv.size / 2);
        
            // Retrieve the neighbour location
            const neighbourLoc = chest_entity.getDynamicProperty('neighbour');
        
            if (neighbourLoc) {
                // Spawn new single chest entity
                const newEntityType = chest_entity.typeId.replace('_double', '');
                const { x, y, z } = neighbourLoc;
                const summonCommand = `summon ${newEntityType} ${x} ${y} ${z}`;
                dimension.runCommand(summonCommand);
            
                // Find the newly summoned entity
                system.runTimeout(() => {
                    const newChestEntity = Array.from(dimension.getEntities({ location: neighbourLoc, maxDistance: 1 }))
                        .find(entity => {
                            const key = `chest_entity_${entity.id}`;
                            return entity.typeId === newEntityType && !chestEntities.has(key);
                        });
    
                    if (!newChestEntity) {
                        console.warn("Failed to find newly summoned chest entity");
                        return;
                    }
                    newChestEntity.nameTag = newChestEntity.typeId.replace("aga_iron_chest:", "");
                    saveEntityLocation({ id: newChestEntity.id, typeId: newChestEntity.typeId }, neighbourLoc, dimension);
        
                    // Transfer first half of items to new entity
                    const newInv = newChestEntity.getComponent('inventory').container;
                    for (let i = 0; i < halfSize; i++) {
                        const item = inv.getItem(i);
                        if (item) {
                            newInv.addItem(item);
                            inv.setItem(i, undefined);
                        }
                    }
                    const spawnLoc = {x: location.x, y: location.y + 1, z: location.z};
        
                    // Drop remaining items
                    for (let i = halfSize; i < inv.size; i++) {
                        const item = inv.getItem(i);
                        if (item) {
                            dimension.spawnItem(item, spawnLoc);
                            inv.setItem(i, undefined);
                        }
                    }
        
                    // Remove the double chest entity
                    chest_entity.remove();
                    removeEntityLocation(chest_entity);
                }, 1); // Short delay to ensure entity is saved in the map
            } else {
                // Handle the case where we're breaking a single chest next to a double chest
                dropAllItems(chest_entity, location);
                chest_entity.remove();
                removeEntityLocation(chest_entity);
            }
        } else {
            // Handle single chest drop logic
            dropAllItems(chest_entity, location);
            chest_entity.remove();
            removeEntityLocation(chest_entity);
        }
    }

    getChest(entity) {
        const { location, dimension } = entity;
        const chest = dimension.getBlock(location);
        if (chest?.hasTag('custom_chest') || chest?.hasTag('custom_barrel'))
            return chest;
    }
    connect() {
        const { chest_entity } = this;
        if (!chest_entity || !chest_entity.dimension) {
            return;
        }
    
        const { dimension } = chest_entity;
        const block = dimension.getBlock(chest_entity.location);
    
        if (!block) {
            return;
        }
    
        // Check if the chest is already part of a double chest
        const currentDoubleState = block.permutation.getState('aga_iron_chest:top') || block.permutation.getState('aga_iron_chest:bottom');
        if (currentDoubleState === 1) {
            return;
        }
    
        // Get the cardinal direction of the placed chest
        const chestDirection = block.permutation.getState('minecraft:cardinal_direction');
    
        const directions = [
            { dy: 1, direction: 'up' },
            { dy: -1, direction: 'down' }
        ];
    
        for (const dir of directions) {
            const adjacentLocation = { 
                x: block.location.x, 
                y: block.location.y + dir.dy, 
                z: block.location.z 
            };
            const adjacentBlock = dimension.getBlock(adjacentLocation);
    
            if (adjacentBlock && adjacentBlock.typeId === block.typeId) {
                const existingAdjacentState = adjacentBlock.permutation.getState('aga_iron_chest:top') || adjacentBlock.permutation.getState('aga_iron_chest:bottom');
                if (existingAdjacentState === 1) continue;
    
                let newChestState, newAdjacentState, bottomChestLocation;
    
                // Determine top and bottom based on the direction
                if (dir.direction === 'up') {
                    newChestState = 'aga_iron_chest:bottom';
                    newAdjacentState = 'aga_iron_chest:top';
                    bottomChestLocation = block.location;
                } else {
                    newChestState = 'aga_iron_chest:top';
                    newAdjacentState = 'aga_iron_chest:bottom';
                    bottomChestLocation = adjacentLocation;
                }
    
                // Set the double chest states
                block.setPermutation(block.permutation.withState(newChestState, 1));
                adjacentBlock.setPermutation(adjacentBlock.permutation.withState(newAdjacentState, 1));
    
                // Convert to double chest
                const doubleChestEntity = this.convertToDoubleChest(chest_entity, adjacentBlock, bottomChestLocation);
    
                if (doubleChestEntity) {
                    doubleChestEntity.setDynamicProperty('doubleDirection', dir.direction);
                    doubleChestEntity.setDynamicProperty('cardinalDirection', chestDirection);
                }
    
                break;
            }
        }
    }
    
    convertToDoubleChest(chest_entity, adjacentBlock, bottomChestLocation) {
        const chest = this.getChest(chest_entity);
        if (!chest) return;
    
        const { dimension } = chest_entity;
    
        // Find the adjacent chest entity
        let adjacentEntity = null;
        const nearbyEntities = Array.from(dimension.getEntities({ location: adjacentBlock.location, maxDistance: 1 }));
        for (const entity of nearbyEntities) {
            if (entity.typeId.includes('chest_entity')) {
                const entityData = chestEntities.get(`chest_entity_${entity.id}`);
                if (entityData && 
                    entityData.x === adjacentBlock.location.x && 
                    entityData.y === adjacentBlock.location.y && 
                    entityData.z === adjacentBlock.location.z) {
                    adjacentEntity = entity;
                    break;
                }
            }
        }
    
        if (!adjacentEntity) {
            console.warn("Failed to find adjacent chest entity");
            return;
        }
    
        // Determine the new double chest entity type
        const newEntityType = chest_entity.typeId + '_double';
    
        // Summon the new double chest entity using runCommand at the bottom chest location
        const { x, y, z } = bottomChestLocation;
        const summonCommand = `summon ${newEntityType} ${x} ${y} ${z}`;
        dimension.runCommand(summonCommand);
    
        // Find the newly summoned entity
        const newEntity = Array.from(dimension.getEntities({ location: bottomChestLocation, maxDistance: 1 }))
            .find(entity => entity.typeId === newEntityType && !chestEntities.has(`chest_entity_${entity.id}`));
    
        if (!newEntity) {
            console.warn("Failed to summon new double chest entity");
            return;
        }
        newEntity.nameTag = newEntity.typeId.replace("aga_iron_chest:", "");
    
        // Transfer items from both chests to the new double chest
        this.transferItems(chest_entity, newEntity);
        this.transferItems(adjacentEntity, newEntity);
    
        // Remove the old chest entities
        chest_entity.remove();
        adjacentEntity.remove();
    
        // Save the new double chest entity location
        saveEntityLocation({ id: newEntity.id, typeId: newEntity.typeId }, bottomChestLocation, newEntity.dimension);
    
        return newEntity;
    }

    transferItems(fromEntity, toEntity) {
        const fromInventory = fromEntity.getComponent('inventory').container;
        const toInventory = toEntity.getComponent('inventory').container;

        for (let i = 0; i < fromInventory.size; i++) {
            const item = fromInventory.getItem(i);
            if (item) {
                toInventory.addItem(item);
                fromInventory.setItem(i, undefined);
            }
        }
    }
}

    function dropAllItems(entity, location) {
        const inv = entity.getComponent('inventory').container;
        for (let i = 0; i < inv.size; i++) {
            const item = inv.getItem(i);
            if (item) {
                entity.dimension.spawnItem(item, location);
                inv.setItem(i, undefined);
            }
        }
    }
