import { world, system, BlockPermutation } from "@minecraft/server";
import { ChestUtils } from "ChestUtils";

const CHEST_ENTITY_DATA_KEY = "aga_iron_chest:chest_entity_data";
const chestEntities = new Map();
const EntityMap = new Map([
    ["aga_iron_chest:copper_chest", "aga_iron_chest:chest_entity_copper"],
    ["aga_iron_chest:dirt_chest", "aga_iron_chest:chest_entity_dirt"],
    ["aga_iron_chest:crystal_chest", "aga_iron_chest:chest_entity_crystal"],
    ["aga_iron_chest:diamond_chest", "aga_iron_chest:chest_entity_diamond"],
    ["aga_iron_chest:iron_chest", "aga_iron_chest:chest_entity_iron"],
    ["aga_iron_chest:golden_chest", "aga_iron_chest:chest_entity_golden"],
    ["aga_iron_chest:netherite_chest", "aga_iron_chest:chest_entity_netherite"],
    ["aga_iron_chest:obsidian_chest", "aga_iron_chest:chest_entity_obsidian"],["aga_iron_chest:iron_shulker_box", "aga_iron_chest:shulker_entity_iron"]
]);

world.beforeEvents.worldInitialize.subscribe(({ blockComponentRegistry }) => {
      blockComponentRegistry.registerCustomComponent("aga_iron_chest:break_chest", breakChest);
  });

world.beforeEvents.playerInteractWithEntity.subscribe((e) => {
    const { target: entity, player } = e;
    if (player.isSneaking) {
        e.cancel = true;
        const equipment = player.getComponent('equippable');
        const selectedItem = equipment.getEquipment('Mainhand');
        if (!selectedItem || selectedItem === undefined) return;
        if (selectedItem?.typeId === 'minecraft:hopper') {
            // Get the block at the entity's location
            const chestBlock = player.dimension.getBlock(entity.location);
            
            if (chestBlock) {
                // Determine the facing direction based on the player's position relative to the entity
                const facingDirection = getFacingDirection(player, entity);
                
                // Calculate the position for the hopper
                const hopperLocation = getAdjacentBlockLocation(chestBlock.location, facingDirection);
                const hopperBlock = player.dimension.getBlock(hopperLocation);

                // Check if the block is air and no entities are present
                if (hopperBlock.typeId === 'minecraft:air' && !hasEntitiesAt(player.dimension, hopperLocation)) {
                    // Create a new hopper block with the correct facing direction
                    const hopperPermutation = BlockPermutation.resolve('minecraft:hopper')
                        .withState('facing_direction', facingDirection);
                    
                    // Set the block to the new hopper
                    system.run(() => {
                        hopperBlock.setPermutation(hopperPermutation);

                    if (player.getGameMode() !== "creative") {
                        if (selectedItem.amount === 1) {
                            equipment.setEquipment('Mainhand', undefined);
                        } else {
                            selectedItem.amount -= 1;
                            equipment.setEquipment('Mainhand', selectedItem);
                        }
                    }});
                } 
            }
        }
    }
    if (!entity.typeId.includes("aga_iron_chest:chest_entity"))
        return;
    system.run(() => {
        new ChestUtils(entity).open();
    });
});

world.afterEvents.playerPlaceBlock.subscribe((e) => {
    const block = e.block;
    const { dimension, location: loc } = block;
    
    if (block.hasTag("custom_chest") || block.hasTag("custom_barrel")) {
        const placedItemType = block.typeId;
        const entityTypeId = EntityMap.get(placedItemType);
        
        if (entityTypeId) {
            const { x, y, z } = loc;
            const summonCommand = `summon ${entityTypeId} ${x} ${y} ${z}`;
            dimension.runCommand(summonCommand);

            // Find the newly summoned entity
            system.runTimeout(() => {
                const newChestEntity = Array.from(dimension.getEntities({ location: loc, maxDistance: 1 }))
                    .find(entity => {
                        const key = `chest_entity_${entity.id}`;
                        return entity.typeId === entityTypeId && !chestEntities.has(key);
                    });

                if (!newChestEntity) {
                    console.warn("Failed to find newly summoned chest entity");
                    return;
                }

                newChestEntity.nameTag = newChestEntity.typeId.replace("aga_iron_chest:", "");
                saveEntityLocation({ id: newChestEntity.id, typeId: newChestEntity.typeId }, loc, dimension);
                
                if (block.hasTag("custom_chest")) {
                    try {
                        new ChestUtils(newChestEntity).connect();
                    } catch (error) {
                        console.warn(`Error connecting chest: ${error}`);
                    }
                }
            }, 1); // Run after a 1 tick delay to ensure the entity is fully spawned
        } 
    }
});

world.afterEvents.worldInitialize.subscribe(() => {
    system.runTimeout(() => {
        updateChestEntitiesMap();
    }, 100); // 5-second delay (100 ticks)
});

world.afterEvents.entityRemove.subscribe((event) => {
    if (event.entity && (event.entity.typeId.includes("aga_iron_chest:chest_entity"))) {
        removeEntityLocation(event.entity);
    }
});

world.afterEvents.blockExplode.subscribe((event) => {
    const { explodedBlockPermutation: perm, dimension, block } = event;
    const loc = block.location;
    if (perm.type.id.includes(`aga_iron_chest:`)) {
        handleChestExplosion(perm, dimension, loc);
    }
});

const breakChest = {
    onPlayerDestroy(e) {
        const { block, player, dimension, destroyedBlockPermutation: perm } = e;
        const loc = block.location;
        const topHalf = perm.getState(`aga_iron_chest:top`);
        const bottomHalf = perm.getState(`aga_iron_chest:bottom`);

        // Function to process the chest entity
        const processChestEntity = (entity, neighbourLoc) => {
            entity.setDynamicProperty('neighbour', block.below());
            system.runTimeout(() => {
                new ChestUtils(entity).drop();
            }, 1);
            const remainingChest = dimension.getBlock(block.below());
            if (remainingChest) {
                const singleChestPermutation = remainingChest.permutation
                    .withState('aga_iron_chest:top', 0)
                    .withState('aga_iron_chest:bottom', 0);
                remainingChest.setPermutation(singleChestPermutation);
            }
        };

        // Check both the current location and the neighbour location
        const checkLocations = [loc];
        if (topHalf !== 0 || bottomHalf !== 0) {
            const neighbourString = getNeighbour(loc, topHalf !== 0);
            const [x, y, z] = neighbourString.split(',').map(Number);
            const neighbourLoc = { x, y, z };
            checkLocations.push(neighbourLoc);
        }

        let chestEntityFound = false;
        for (const checkLoc of checkLocations) {
            const entities = dimension.getEntities({
                location: checkLoc,
                maxDistance: 1
            });

            for (const entity of entities) {
                if (entity && typeof entity.typeId === 'string' && entity.typeId.includes("aga_iron_chest:chest_entity")) {
                    const entityData = chestEntities.get(`chest_entity_${entity.id}`);
                    if (entityData && 
                        entityData.x === Math.floor(checkLoc.x) && 
                        entityData.y === Math.floor(checkLoc.y) && 
                        entityData.z === Math.floor(checkLoc.z)) {
                        chestEntityFound = true;
                        if (topHalf !== 0 || bottomHalf !== 0) {
                            const neighbourString = getNeighbour(loc, topHalf !== 0);
                            const [x, y, z] = neighbourString.split(',').map(Number);
                            const neighbourLoc = { x, y, z };
                            processChestEntity(entity, neighbourLoc);
                        } else {
                            new ChestUtils(entity).drop();
                        }
                        break;
                    }
                }
            }

            if (chestEntityFound) break;
        }

        if (player.getGameMode() !== "creative") {
            const inventory = player.getComponent("inventory");
            const selectedItem = inventory.container.getItem(player.selectedSlotIndex);
            if (
                selectedItem &&
                (selectedItem.hasTag("minecraft:is_shovel") ||
                selectedItem.hasTag("minecraft:is_hoe") ||
                selectedItem.hasTag("minecraft:is_axe") ||
                selectedItem.hasTag("minecraft:is_pickaxe") ||
                selectedItem.hasTag("minecraft:is_shears"))
            ) {
                system.runTimeout(() => {
                    applyDurabilityDamage(player, selectedItem, inventory, player.selectedSlotIndex);
                }, 1);
            } 
        }
    },
};

function getFacingDirection(player, entity) {
    const dx = player.location.x - entity.location.x;
    const dz = player.location.z - entity.location.z;
    
    if (Math.abs(dx) > Math.abs(dz)) {
        return dx > 0 ? 4 : 5; // 5 for west, 4 for east
    } else {
        return dz > 0 ? 2 : 3; // 3 for north, 2 for south
    }
}

function getAdjacentBlockLocation(location, direction) {
    switch (direction) {
        case 2: return { x: location.x, y: location.y, z: location.z + 1 }; // South
        case 3: return { x: location.x, y: location.y, z: location.z - 1 }; // North
        case 4: return { x: location.x + 1, y: location.y, z: location.z }; // East
        case 5: return { x: location.x - 1, y: location.y, z: location.z }; // West
        default: return location;
    }
}

function hasEntitiesAt(dimension, location) {
    const entities = dimension.getEntities({ location: location, maxDistance: 1 });
    return entities.some(entity => !entity.typeId.includes("chest_entity"));
}

  function getNeighbour(loc, isTop) {
    const sign = isTop ? -1 : 1;
    let x = loc.x, y = loc.y += sign, z = loc.z;
    if (sign || !sign) {
        return `${x}, ${y}, ${z}`;
    }
}

function updateChestEntitiesMap() {
    chestEntities.clear();
    const data = world.getDynamicProperty(CHEST_ENTITY_DATA_KEY);
    if (data) {
        try {
            const parsedData = JSON.parse(data);
            for (const [key, value] of Object.entries(parsedData)) {
                chestEntities.set(key, value);
            }
        } catch (error) {
            console.warn(`Error parsing chest entity data: ${error}`);
        }
    } 
}

export function saveEntityLocation(entityInfo, location, dimension) {
    
    if (entityInfo && entityInfo.id && entityInfo.typeId) {
        const key = `chest_entity_${entityInfo.id}`;
        const value = {
            type: entityInfo.typeId,
            id: entityInfo.id,
            x: location.x,
            y: location.y,
            z: location.z,
            dimension: dimension.id,
            isOpen: false // Default to closed
        };

        chestEntities.set(key, value);
        saveChestEntityData();
    } 
}

export function removeEntityLocation(entity) {
    const key = `chest_entity_${entity.id}`;
    
    // Get the current chest entity data
    const chestEntityData = world.getDynamicProperty("aga_iron_chest:chest_entity_data");
    
    if (chestEntityData) {
        try {
            const parsedData = JSON.parse(chestEntityData);
            
            // Remove the specific entity from the data
            if (parsedData[key]) {
                delete parsedData[key];
                
                // Save the updated data back to the dynamic property
                world.setDynamicProperty("aga_iron_chest:chest_entity_data", JSON.stringify(parsedData));
                
                // Update the chestEntities map
                chestEntities.delete(key);
            } 
        } catch (error) {
            console.error(`Error removing entity data for ${key}:`, error);
        }
    } 
}

function saveChestEntityData() {
    const data = Object.fromEntries(chestEntities);
    world.setDynamicProperty(CHEST_ENTITY_DATA_KEY, JSON.stringify(data));
}

function applyDurabilityDamage(player, item, inventory, slotIndex) {
  const durabilityComponent = item.getComponent("minecraft:durability");
  if (durabilityComponent) {
    const { unbreakingLevel } = getRelevantEnchantments(item);
    
    if (Math.random() < 1 / (unbreakingLevel + 1)) {
      const newDamage = durabilityComponent.damage + 1;
      if (newDamage >= durabilityComponent.maxDurability) {
        inventory.container.setItem(slotIndex, undefined);
        player.playSound("random.break");
      } else {
        durabilityComponent.damage = newDamage;
        inventory.container.setItem(slotIndex, item);
      }
    }
  }
}

function getRelevantEnchantments(item) {
  let unbreakingLevel = 0;

  try {
      const enchantableComponent = item.getComponent("minecraft:enchantable");
      if (enchantableComponent) {
          const enchantments = enchantableComponent.getEnchantments();
          for (const enchant of enchantments) {
              if (enchant.type.id === "unbreaking") {
                  unbreakingLevel = enchant.level;
              }
          }
      }
  } catch (error) {
  }
  return { unbreakingLevel };
}

function handleChestExplosion(perm, dimension, loc) {
    const topHalf = perm.getState(`aga_iron_chest:top`);
    const bottomHalf = perm.getState(`aga_iron_chest:bottom`);
    const cardinal = perm.getState(`minecraft:cardinal_direction`);

    const blockBelow = dimension.getBlock({x: loc.x, y: loc.y -1, z: loc.z})

    // Function to process the chest entity
    const processChestEntity = (entity, neighbourLoc) => {
        entity.setDynamicProperty('neighbour', blockBelow);
        system.runTimeout(() => {
            new ChestUtils(entity).drop();
        }, 1);
        const remainingChest = dimension.getBlock(blockBelow);
        if (remainingChest && remainingChest.typeId.includes(`aga_iron_chest:`)) {
            const singleChestPermutation = remainingChest.permutation
                .withState('aga_iron_chest:top', 0)
                .withState('aga_iron_chest:bottom', 0);
            remainingChest.setPermutation(singleChestPermutation);
        }
    };

    // Check both the current location and the neighbour location
    const checkLocations = [loc];
    if (topHalf !== 0 || bottomHalf !== 0) {
        const neighbourString = getNeighbour(loc, topHalf !== 0);
        const [x, y, z] = neighbourString.split(',').map(Number);
        const neighbourLoc = { x, y, z };
        checkLocations.push(neighbourLoc);
    }

        let chestEntityFound = false;
        for (const checkLoc of checkLocations) {
            const entities = dimension.getEntities({
                location: checkLoc,
                maxDistance: 1
            });

            for (const entity of entities) {
                if (entity && typeof entity.typeId === 'string' && entity.typeId.includes("aga_iron_chest:chest_entity")) {
                    const entityData = chestEntities.get(`chest_entity_${entity.id}`);
                    if (entityData && 
                        entityData.x === Math.floor(checkLoc.x) && 
                        entityData.y === Math.floor(checkLoc.y) && 
                        entityData.z === Math.floor(checkLoc.z)) {
                        chestEntityFound = true;
                        if (topHalf !== 0 || bottomHalf !== 0) {
                            const neighbourString = getNeighbour(loc, topHalf !== 0);
                            const [x, y, z] = neighbourString.split(',').map(Number);
                            const neighbourLoc = { x, y, z };
                            processChestEntity(entity, neighbourLoc);
                        } else {
                            new ChestUtils(entity).drop();
                        }
                        break;
                    }
                }
            }

            if (chestEntityFound) break;
        }

    // If no chest entity was found, we might need to handle the case where both parts of a double chest are destroyed simultaneously
    if (!chestEntityFound && (topHalf !== 0 || bottomHalf !== 0)) {
        const neighbourString = getNeighbour(loc, cardinal, topHalf !== 0);
        const [x, y, z] = neighbourString.split(',').map(Number);
        const neighbourLoc = { x, y, z };
        const neighbourEntities = dimension.getEntities({
            location: neighbourLoc,
            maxDistance: 1
        });

        for (const entity of neighbourEntities) {
            if (entity && typeof entity.typeId === 'string' && entity.typeId.includes("aga_iron_chest:chest_entity")) {
                new ChestUtils(entity).drop();
                break;
            }
        }
    }
}

export {
    chestEntities
}